//
//  Cube4DView.m
//  Cube4D
//
//  Created by Alex Boreskoff on 4/16/06.
//  Copyright (c) 2006, __MyCompanyName__. All rights reserved.
//

#import "Cube4DView.h"
#import <math.h>

static	Vector4D e1 ( 1, 2, -5, 9 );
static	Vector4D e2 ( -9, 4, 2, 7 );
static	Vector4D dir ( 1, -1, 3, -1 );

static	float		angle1  = 0;
static	float		angle2  = 0;
static	float		step1   = 0.04f;
static	float		step2   = 0.01f;

void	orthonormalize ()
{
	float	dot    = e1 & e2;
	float	len1Sq = e1 & e1;
	
	e2 -= e1 * dot / len1Sq;
	
	e1.normalize ();
	e2.normalize ();
}

void	drawCube ( NSRect bounds, int brightness, float a1, float a2 )
{
	NSPoint		center;
	NSSize		scale;
	float		matrix [4][4];
	int			i, j;
	float		br      = (float) brightness / 255.0f;
	NSBezierPath * path = [NSBezierPath bezierPath];
	
	[[NSColor colorWithDeviceRed: br green: br blue: br alpha: 1.0] set];
	
	center.x     = bounds.origin.x + bounds.size.width  * 0.5;
	center.y     = bounds.origin.y + bounds.size.height * 0.5;
	scale.width  = bounds.size.width  / 45.0;
	scale.height = bounds.size.height / 45.0;
	
	for ( i = 0; i < 4; i++ )
		for ( j = 0; j < 4; j++ )
			matrix [i][j] = 0;
	
	matrix [0][0] =  (float) cos ( a1 );
	matrix [1][1] =  (float) cos ( a1 );
	matrix [0][1] =  (float) sin ( a1 );
	matrix [1][0] = -(float) sin ( a1 );
	matrix [2][2] =  (float) cos ( a2 );
	matrix [3][3] =  (float) cos ( a2 );
	matrix [2][3] =  (float) sin ( a2 );
	matrix [3][2] = -(float) sin ( a2 );
	
	Vector4D	b1, b2;
	
	for ( i = 0; i < 4; i++ )
	{
		b1 [i] = 0;
		b2 [i] = 0;
		
		for ( int j = 0; j < 4; j++ )
		{
			b1 [i] += e1 [j] * matrix [i][j];
			b2 [i] += e2 [j] * matrix [i][j];
		}
	}
	
	for ( i = 0; i < 15; i++ )
	{
		Vector4D	v;
		
								// fill v with binary representation of i
		v.w = (i & 1 ? 1.0f : 0.0f);
		v.z = (i & 2 ? 1.0f : 0.0f);
		v.y = (i & 4 ? 1.0f : 0.0f);
		v.x = (i & 8 ? 1.0f : 0.0f);
		
		for ( j = 0; j < 4; j++ )
		{
			Vector4D	   v2 = v;
			NSPoint		   start, end;
			
			if ( v [j] == 1 )
				continue;
			
			v2 [j] = 1.0;
			
			Vector4D	offs ( sin ( angle1 * 0.33f + angle2 * 0.2f ) * dir / 4 );
			Vector4D	w  = v  + offs;
			Vector4D	w2 = v2 + offs;
			
			start.x = ( w  &  b1 ) / (0.7 + 0.2*v.w);
			start.y = ( w  &  b2 ) / (0.7 + 0.2*v.w);
			end.x   = ( w2 & b1  ) / (0.7 + 0.2*v2.w);
			end.y   = ( w2 & b2  ) / (0.7 + 0.2*v2.w);
			
			// remap projection to bounds
			start.x = center.x + start.x * scale.width;
			start.y = center.y + start.y * scale.height;
			end.x   = center.x + end.x   * scale.width;
			end.y   = center.y + end.y   * scale.height;
			
			[path moveToPoint: start];
			[path lineToPoint: end];
		}	
	}
	
	[path closePath];
	[path stroke];
}

@implementation Cube4DView

static	NSString * cube4DModuleName = @"com.alex.Cube4D";

- (id)initWithFrame: (NSRect)frame isPreview:(BOOL) isPreview
{
    self = [super initWithFrame:frame isPreview:isPreview];
	
    if (self)
	{
		ScreenSaverDefaults * defs;
		
		defs = [ScreenSaverDefaults defaultsForModuleWithName: cube4DModuleName];
		
								// register default values
		[defs registerDefaults:[NSDictionary dictionaryWithObjectsAndKeys: 
			[NSNumber numberWithInt: 3], @"ShadowSize", nil]];
		
		shadowSizeValue = [defs integerForKey: @"ShadowSize"];
		
        [self setAnimationTimeInterval:1/20.0];
	}
	
    return self;
}

- (BOOL) hasConfigureSheet
{
	return YES;
}

- (NSWindow *) configureSheet
{
	ScreenSaverDefaults * defs;
	
	if ( !configSheet )
	{
		if ( ! [NSBundle loadNibNamed: @"Cube4D" owner: self] )
		{
			NSBeep ();
		}
	}
	
	defs = [ScreenSaverDefaults defaultsForModuleWithName: cube4DModuleName];
	
	[shadowSize setIntValue: shadowSizeValue];
	
	return configSheet;
}

- (void) startAnimation
{
    [super startAnimation];
}

- (void) stopAnimation
{
    [super stopAnimation];
}

- (void) drawRect:(NSRect)rect
{
	int	i;
	
    [super drawRect:rect];
	
	[[NSColor blackColor] set];
	[[NSBezierPath bezierPathWithRect: rect] fill];
	
	for ( i = 1 - shadowSizeValue; i <= 0; i++ )
	{
		int	intensity = 255 + (255*i) / shadowSizeValue;
		
		drawCube ( rect, intensity,  angle1 + i * step1 / 2, angle2 + i * step2 / 2 );	
	}
	/*	
		drawCube ( rect, 64,  angle1 - 4 * step1 / 2, angle2 - 4 * step2 / 2 );
	 drawCube ( rect, 96,  angle1 - 3 * step1 / 2, angle2 - 3 * step2 / 2 );
	 drawCube ( rect, 128, angle1 - 2 * step1 / 2, angle2 - 2 * step2 / 2 );
	 drawCube ( rect, 192, angle1 -     step1 / 2, angle2 -     step2 );
	 drawCube ( rect, 255, angle1,                 angle2 );
	 */	
	angle1 += step1;
	angle2 += step2;
}

- (void)animateOneFrame
{
	[self setNeedsDisplay: YES];
	
    return;
}

- (IBAction) cancelClick: (id) sender
{
	[[NSApplication sharedApplication] endSheet: configSheet];
}

- (IBAction) okClick: (id) sender
{
	ScreenSaverDefaults * defs;
	
	defs            = [ScreenSaverDefaults defaultsForModuleWithName: cube4DModuleName];
	shadowSizeValue = [shadowSize intValue];
	
	[defs setInteger: shadowSizeValue forKey: @"ShadowSize"];
	[defs synchronize];
	
	[[NSApplication sharedApplication] endSheet: configSheet];
}

@end
